///////////////////////////////////////////////////////////////////////////
// BB20 - Commodore Vic-20 Emulator for Blackberry 957                   //
// This version uses Neil Bradley's M6502 Assembly Language 6502 Core    //
// Basic: C000-DFFF  Kernal:E000-FFFF  Screen:1E00-1FF9  ChGen:XXXX-XXXX //
///////////////////////////////////////////////////////////////////////////
#include <pager.h>
#include <ribbon.h>
#include <string.h>
#include "icon.h"
#include "M6502.h" // M6502 structures
#include "fnt.h"   // Vic20 Font
#include "rom.h"   // Combined Basic & Kernal roms

char VersionPtr[] = "BB20";
int AppStackSize = 1024;

CONTEXTM6502 *c20;  // Cpu context
unsigned char *mem; // The memory
UINT32 dwElapsedTicks = 0;
UINT32 cyclesRemaining = 0;

inline void screen(UINT32 scr, UINT8 chr, struct MemoryWriteByte *psMW) {
  register int k = scr-0x1e00;
  mem[scr] = chr;
  LcdPutStringXY(3+(k%22)*7,(k/22)*7, &chr, 1, LCD_RAW);
}

inline void vicsnd(UINT32 add, UINT8 dat, struct MemoryWriteByte *psMW) {
  RimSpeakerBeep(dat, 70, 50, (mem[0x900E]&15)*6); // Play snd @ cur vol
}

inline void vicvib(UINT32 add, UINT8 dat, struct MemoryWriteByte *psMW) {
  AlertConfiguration VIB = {50, 1, 1, 1, dat?4:0, 1, 2, 2, 0};
  RimTestAlert(1, &VIB);
}

void fileio(UINT32 adr, UINT8 val, struct MemoryWriteByte *psMW) {
  m6502zpGetContext(c20);
  if (c20->m6502pc == 0xE15F || c20->m6502pc == 0xE175) {
    unsigned int opA, opB;    // Tmps
    unsigned char lo, hi;     // Load address
    HandleType H; FileType F; // Load/Save types
    char *z = strchr(&mem[0x0202], '"'); *z = '\0';
    H=DbGetHandle(&mem[0x0202]); F=DbFileOpen(H,256);
    if (c20->m6502pc == 0xE175) {                   // LOAD
      DbFileRead(F,&lo,1); DbFileRead(F,&hi,1);     // Load address
      opA = DbSize(H,TRUE); opB = hi*256+lo;        // Get file len & start loc
      DbFileRead(F,&mem[opB],opA);                  // File content
      if (lo == mem[0x002b] && hi == mem[0x002c]) { // Normal basic prg
        adr = opB + opA -2;
        c20->m6502x = adr&255; c20->m6502y = adr>>8; c20->m6502af = 0xC7; // c7=11000111
      }
    } else {                                        // SAVE
      DbFileWrite(F,&mem[0x002B],1); DbFileWrite(F,&mem[0x002C],1);     // Load address
      DbFileWrite(F,&mem[0x1001],(mem[0x002e]*256+mem[0x002d])-0x1001); // File content
      c20->m6502af = 0;
    }
    DbFileClose(F,TRUE);
    RimSpeakerBeep(110,50,50,50);
    m6502zpSetContext(c20);
  }
}

struct MemoryWriteByte write20[] = {
  {0x1E00, 0x1FF9, screen}, // Screen: 1E00-1FF9
  {0x0313, 0x0313, fileio}, // LOAD (FFD5) or SAVE (FFD8)
  {0x900A, 0x900A, vicsnd}, // Basic 1-voice sound (osc 1)
  {0x900D, 0x900D, vicvib}, // Noise gen to vibrate
  {(UINT32) - 1, (UINT32) - 1, NULL}
};

struct MemoryReadByte read20[] = {
  {(UINT32) - 1, (UINT32) - 1, NULL}
};

int kFix(int k) {
  if (k == KEY_DELETE) { // NOT WORKING TRY IO chip
    k = 0;
    mem[0x912F] = 0xFE;       // Stop key pressed
  } else mem[0x912F] = 0xFF;  // Stop key not pressed
//mem[0x91] = (k == KEY_DELETE)? 127 : 0;
  if (k > 96 && k < 123)   return k - 32;
  if (k > 64 && k < 91)    return k + 128;
  if (k == KEY_ENTER)      return 13; // CR
  if (k == KEY_BACKSPACE)  return 20; // Delete
  return k & 0xff;
}

void kPut(int k) {
  if (mem[198]<10)  mem[631+mem[198]++] = k;
}

void BB20Thread(void) {
//#include "fnt.h"   // Vic20 Font
//#include "rom.h"   // Combined Basic & Kernal roms
  MESSAGE M;
  PID pid;
  pid.Name = VersionPtr;
  pid.EnableForeground = TRUE;
  pid.Icon = &IconBMP;
  RimSetPID(&pid);
  RimRequestForeground(RimGetCurrentTaskID());
  LcdReplaceFont(2,&cbm1);
  LcdSetCurrentFont(2);                  // Switch to Vic-20 22x23 font
  mem=(unsigned char *)RimMalloc(65536); // Request 64k of memory
  memset(mem, 0, 65536);                 // Clr mem because mem test is skipped
  memcpy(&mem[0xC000], rom, 16384);      // Load basic & kernal roms
  // Rom mods       sMem = 0x1e00; sBas = 0x1000; eBas = 0x1e00
  mem[0xFD2F] = 0xEA; mem[0xFD30] = 0xEA; mem[0xFD31] = 0xEA;  // Skip mem test
  mem[0x00b2] = 0x3c; mem[0x00b3] = 0x03;  // Pointer: start of tape buffer (033c)
  mem[0x0288] = 0x1e; // Set Screen memory
  mem[0x0282] = 0x10; // Set Start of Basic
  mem[0x00c2] = 0x1e; // Set 
  mem[0x0284] = 0x1e; // Set End of Basic
  mem[0xEACB] = 0x04; // Cursor speedup
  // Hack: Change JSR to STA 0313 to trigger exception for Load and Save
  mem[0xE15C] = 0x8D; mem[0xE15D] = 0x13; mem[0xE15E] = 0x03; // SAVE @ E15C
  mem[0xE172] = 0x8D; mem[0xE173] = 0x13; mem[0xE174] = 0x03; // LOAD @ E172
  c20 = RimMalloc(m6502zpGetContextSize());
  memset(c20, 0, m6502zpGetContextSize()); // Clear cpu context
  c20->m6502Base = mem;
  c20->m6502MemoryRead = read20;
  c20->m6502MemoryWrite = write20;
  m6502zpSetContext(c20); // Set the context of the 6502 emulator core
  m6502zpreset();         // Reset the processor to get the start vector

  for (;;) {
    while (RimPeekMessage() == FALSE) {
      m6502zpexec(32000);
      m6502zpint(0);
      mem[0x9114]++; // Rnd(0) fix, mix it up a little
      RimSleep(1);   // 10ms, min req to reset watchdog timer
    }
    RimGetMessage(&M);
    if (M.Device == DEVICE_KEYPAD && M.Event == THUMB_ROLL_UP)
      kPut(M.Data[0] & ALT_STATUS?29:145);  // Emulate Cursor keys R or U
    if (M.Device == DEVICE_KEYPAD && M.Event == THUMB_ROLL_DOWN)
      kPut(M.Data[0] & ALT_STATUS?157:17);  // Emulate Cursor keys L or D
    if (M.Device == DEVICE_KEYPAD && (M.Event == KEY_DOWN || M.Event == KEY_REPEAT))
      if (M.SubMsg == KEY_ESCAPE) {
        RimFree(mem);       // Release Vic memory
        RimFree(c20);       // Release cpu context
        RibbonShowRibbon();
        RimTerminateThread();
       } else kPut(kFix(M.SubMsg)); // Key pressed
   } // End for (;;)
}

void PagerMain(void) {
  RibbonRegisterFunction(VersionPtr, &IconBMP, 0, 40, BB20Thread, AppStackSize);
}
